#include <stdbool.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <sha33/sha33.h>

#include "drup/types.h"
#include "private/auth.h"
#include "private/ds1961.h"
#include "private/ds2482.h"
#include "private/util.h"

#define NUM_RETRIES_WRITE_SCRATCHPAD 10
#define NUM_RETRIES_READ_AUTH_PAGE 10

void make_challenge (uint8_t * const buf) {
  srandom (time (NULL));
  for (int i = 0; i < 3; i++)
    buf [i] = random ();
}

/* Doesn't always set errno.
 */
static bool read_with_challenge (const uint8_t cfg, const struct key * const key, const uint8_t * const challenge, uint8_t * const auth_page_read, uint8_t * const rmac) {
  // --- 10ms
  static const struct timespec pass_wait = { .tv_sec = 0, .tv_nsec = 10e6, };
  // --- 100ms
  static const struct timespec pass_wait2 = { .tv_sec = 0, .tv_nsec = 1e8, };

  uint8_t data[8] = {
    0, 0, 0, 0, challenge[0], challenge[1], challenge[2], 0,
  };
  bool ok;

  ok = false;
  for (int loop = 0; loop < NUM_RETRIES_WRITE_SCRATCHPAD; loop++) {
    if (ds1961_write_scratchpad (cfg, key, key->auth_addr, data)) {
      ok = true;
      break;
    }
    ac_warn ("ds1961_write_scratchpad () failed (pass %d/%d)", loop + 1, NUM_RETRIES_READ_AUTH_PAGE);
    nanosleep (&pass_wait2, NULL);
  }
  if (!ok) return false;

  ok = false;
  for (int loop = 0; loop < NUM_RETRIES_READ_AUTH_PAGE; loop++) {
    if (ds1961_read_auth_page (cfg, key, auth_page_read, rmac)) {
      ok = true;
      break;
    }
    ac_warn ("ds1961_read_auth_page () failed (pass %d/%d)", loop + 1, NUM_RETRIES_READ_AUTH_PAGE);
    nanosleep (&pass_wait, NULL);
  }
  if (!ok) return false;

  return true;
}

static void calc_lmac (uint16_t auth_addr, uint64_t device_id, const uint8_t * const challenge, const uint8_t * const secret8, const uint8_t * const auth_page_read, uint8_t * const lmac) {
  uint32_t hash_input[16];
  uint32_t hash_output[16];
  uint8_t id_input[7];

  for (int i = 0; i < 7; i++)
    id_input[i] = (device_id & (0xFFLL << i * 8)) >> i * 8;

  hash_input[0] = (secret8[0] << 24) | (secret8[1] << 16) | (secret8[2] << 8) | secret8[3];

  for (int i = 0; i < 32; i += 4)
    hash_input[i/4 + 1] = (auth_page_read[i] << 24) | (auth_page_read[i + 1] << 16) | (auth_page_read[i + 2] << 8) | auth_page_read[i + 3];

  hash_input[9] = 0xFFFFFFFF;
  uint8_t mp = (0b1000 << 3) | ((auth_addr >> 5) & 0b111);
  hash_input[10] = (mp << 24) | (id_input[0] << 16) | (id_input[1] << 8) | id_input[2];
  hash_input[11] = (id_input[3] << 24) | (id_input[4] << 16) | (id_input[5] << 8) | id_input[6];
  hash_input[12] = (secret8[4] << 24) | (secret8[5] << 16) | (secret8[6] << 8) | secret8[7];
  hash_input[13] = (challenge[0] << 24) | (challenge[1] << 16) | (challenge[2] << 8) | 0x80;
  hash_input[14] = 0;
  hash_input[15] = 0x1B8;

  sha33_compute_sha_vm_32 (hash_input, hash_output);
  sha33_hash_to_mac (hash_output, lmac);
}

/* Doesn't always set errno.
 */
bool auth_check (const struct drup_state * const state, const struct key * const key, uint8_t cfg, bool * const authenticated) {
  uint16_t const auth_addr = key->auth_addr;
  uint64_t const device_id = key->device_id;

  uint8_t challenge[3];
  make_challenge (challenge);

  // --- read during calculation of rmac.
  uint8_t auth_page_read[32];
  // --- calculated by device.
  uint8_t rmac[20];
  // --- calculated by us.
  uint8_t lmac[20];

  try (ds2482_write_configuration, (state->master, cfg))
  try_with (read_with_challenge, (cfg, key, challenge, auth_page_read, rmac), ac_warn_rf)

  uint8_t secret8[8];
  const uint8_t * const secret16 = key->secret;
  secret_16_to_8 (secret16, secret8);
  calc_lmac (auth_addr, device_id, challenge, secret8, auth_page_read, lmac);
  *authenticated = memcmp (lmac, rmac, sizeof rmac) == 0;
  return true;
}
